#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "symbol.h"
#include "hash.h"
#include "xmalloc.h"
#include "mlex.h"

static struct symbol_table *global;
static struct symbol_table *top;

static int symbol_strcmp(void *a, void *b)
{
	int ret;

	return_val_if_fail(a != NULL, ret_invalid);
	return_val_if_fail(b != NULL, ret_invalid);

	ret = strcmp((char *)a, (char *)b);

	return ret;
}

static int symbol_strhash(unsigned char *str)
{
	unsigned long hash = 5381;
	int c;

	while ((c = *str++) != 0)
		hash = ((hash << 5) + hash) + c; /* hash * 33 + c */

	return hash;
}

static const char *get_file_name()
{
	return mlex_file_name();
}

static unsigned int get_line_number()
{
	return mlex_line_number();
}

/* A constructor, destructor and setters of a variable symbol */
static void symbol_variable_release(struct symbol *symbol)
{
	struct symbol_var *var;

	return_if_fail(symbol != NULL);

	var = (struct symbol_var *)symbol;

	xfree(SYMBOL(var)->name);
	xfree(var);
}

struct symbol_var *symbol_variable(const char *name, data_type_t decl_type, storage_class_t klass)
{
	struct symbol_var *var;

	return_val_if_fail(name != NULL, NULL);

	var = xmalloc0(sizeof(*var));

	SYMBOL(var)->type = SYMBOL_TYPE_VARIABLE;
	SYMBOL(var)->name = xstrdup(name);
	SYMBOL(var)->klass = klass;
	SYMBOL(var)->decl_type = decl_type;
	SYMBOL(var)->release = symbol_variable_release;
	SYMBOL(var)->ref_count = 1;

	SYMBOL(var)->file_name = get_file_name();
	SYMBOL(var)->line_number = get_line_number();

	return var;
}

static void symbol_argument_release(struct symbol *symbol)
{
	struct symbol_arg *arg;

	return_if_fail(symbol != NULL);

	arg = (struct symbol_arg *)symbol;
	if (SYMBOL(arg)->name != NULL)
		xfree(SYMBOL(arg)->name);

	if (arg->initval != NULL)
		ast_node_unref(arg->initval);

	xfree(arg);
}

struct symbol_arg *symbol_argument()
{
	struct symbol_arg *arg;

	arg = xmalloc0(sizeof(*arg));

	SYMBOL(arg)->type = SYMBOL_TYPE_ARGUMENT;

	SYMBOL(arg)->klass = SYMBOL_VISIB_UKNOWN;
	SYMBOL(arg)->release = symbol_argument_release;
	SYMBOL(arg)->ref_count = 1;

	SYMBOL(arg)->file_name = get_file_name();
	SYMBOL(arg)->line_number = get_line_number();

	return arg;
}

void symbol_arg_set_name(struct symbol_arg *symbol, const char *name)
{
	return_if_fail(symbol != NULL);
	return_if_fail(name != NULL);

	SYMBOL(symbol)->name = xstrdup(name);
}

void symbol_arg_set_type(struct symbol_arg *symbol, data_type_t type)
{
	return_if_fail(symbol != NULL);

	SYMBOL(symbol)->decl_type = type;
}

void symbol_arg_set_initval(struct symbol_arg *symbol, struct ast_node *initval)
{
	return_if_fail(symbol != NULL);

	symbol->initval = initval;
}

/* A constructor, destructor and setters of an array symbol */
static void symbol_array_release(struct symbol *symbol)
{
	struct symbol_array *array;

	return_if_fail(symbol != NULL);

	array = (struct symbol_array *)symbol;

	xfree(SYMBOL(array)->name);
	if (array->dims != NULL)
		xfree(array->dims);

	xfree(array);
}

struct symbol_array *symbol_array(const char *name, data_type_t decl_type, storage_class_t klass)
{
	struct symbol_array *array;

	return_val_if_fail(name != NULL, NULL);

	array = xmalloc0(sizeof(*array));

	SYMBOL(array)->type = SYMBOL_TYPE_ARRAY;
	SYMBOL(array)->name = xstrdup(name);
	SYMBOL(array)->klass = klass;
	SYMBOL(array)->decl_type = decl_type;
	SYMBOL(array)->release = symbol_array_release;
	SYMBOL(array)->ref_count = 1;

	SYMBOL(array)->file_name = get_file_name();
	SYMBOL(array)->line_number = get_line_number();

	return array;
}

void symbol_array_add_dim(struct symbol_array *symbol, unsigned int size)
{
	return_if_fail(symbol != NULL);

	symbol->ndims++;
	symbol->dims = xrealloc(symbol->dims, sizeof(unsigned int)*symbol->ndims);
	symbol->dims[symbol->ndims-1] = size;
}

/* A constructor, destructor and setters of a function symbol */
static void symbol_func_release(struct symbol *symbol)
{
	struct symbol_func *func;
	unsigned int i;

	return_if_fail(symbol != NULL);

	func = (struct symbol_func *)symbol;

	xfree(SYMBOL(func)->name);

	for (i = 0; i < func->nargs; i++)
		symbol_unref(SYMBOL(func->args[i]));

	if (func->args != NULL)
		xfree(func->args);

	xfree(func);
}

struct symbol_func *symbol_func(const char *name, data_type_t ret_type, storage_class_t klass)
{
	struct symbol_func *func;

	return_val_if_fail(name != NULL, NULL);

	func = xmalloc0(sizeof(*func));

	SYMBOL(func)->type = SYMBOL_TYPE_FUNC;
	SYMBOL(func)->name = xstrdup(name);
	SYMBOL(func)->klass = klass;
	SYMBOL(func)->decl_type = ret_type;
	SYMBOL(func)->release = symbol_func_release;
	SYMBOL(func)->ref_count = 1;

	SYMBOL(func)->file_name = get_file_name();
	SYMBOL(func)->line_number = get_line_number();

	return func;
}

void symbol_func_add_arg(struct symbol_func *func, struct symbol_arg *arg)
{
	return_if_fail(func != NULL);
	return_if_fail(arg != NULL);

	func->nargs++;
	func->args = xrealloc(func->args, sizeof(struct symbol_arg *)*func->nargs);
	func->args[func->nargs-1] = arg;
}

void symbol_func_mark_as_decl(struct symbol_func *func)
{
	return_if_fail(func != NULL);

	SYMBOL(func)->type = SYMBOL_TYPE_FUNC_DECL;
}

/* A compare helper for qsort(). */
static int cmp_symbols(const void *a, const void *b)
{
	struct symbol_var *s_a, *s_b;

	s_a = *(struct symbol_var **)a;
	s_b = *(struct symbol_var **)b;

	if (s_a->scope_num > s_b->scope_num)
		return 1;

	return -1;
}

/* Gathering variables or arguments. */
static struct symbol **extract_variables(struct symbol_table *table, unsigned int *ret_size, boolean_t is_arg)
{
	struct symbol **symbols;
	struct symbol *symbol;
	struct hash_table_iter *iter;
	char *name;
	unsigned int used, size;

	iter = hash_table_iterate_init(table->scope);
	if (iter == NULL)
		xerror(1, "hash iterator");

	symbols = NULL;
	used = size = 0;

	while (hash_table_iterate(iter, (void **)&name, (void **)&symbol)) {
		switch (symbol->type) {
		case SYMBOL_TYPE_ARGUMENT:
			if (!is_arg)
				continue;
		case SYMBOL_TYPE_VARIABLE:
		case SYMBOL_TYPE_ARRAY:
			if (used == size) {
				size += 16;
				symbols = xrealloc(symbols, sizeof(struct symbol *)*size);
			}
			symbols[used++] = symbol;
			break;
		default:
			break;
		}
	}

	hash_table_iterate_deinit(&iter);

	if (symbols != NULL)
		qsort(&symbols[0], used, sizeof(struct symbol_var *), cmp_symbols);

	*ret_size = used;

	return symbols;
}

struct symbol **symbol_table_extract_variables(struct symbol_table *table, unsigned int *size)
{
	return_val_if_fail(table != NULL, NULL);
	return_val_if_fail(size != NULL, NULL);

	return extract_variables(table, size, FALSE);
}

struct symbol **symbol_table_extract_arguments(struct symbol_table *table, unsigned int *size)
{
	return_val_if_fail(table != NULL, NULL);
	return_val_if_fail(size != NULL, NULL);

	return extract_variables(table, size, TRUE);
}

void symbol_incr_reference(struct symbol *symbol)
{
	return_if_fail(symbol != NULL);

	symbol->nr_references++;
}

void symbol_table_put_symbol(struct symbol *symbol)
{
	ret_t ret;

	return_if_fail(symbol != NULL);

	switch (symbol->type) {
	case SYMBOL_TYPE_ARGUMENT:
		SYMBOL_VAR(symbol)->scope_num = top->nargs;
		top->nargs++;
		break;
	case SYMBOL_TYPE_VARIABLE:
	case SYMBOL_TYPE_ARRAY:
		SYMBOL_VAR(symbol)->scope_num = top->nvars;
		top->nvars++;
		break;
	default:
		break;
	}

	ret = hash_table_insert_unique(top->scope, SYMBOL(symbol)->name, symbol);
	if (ret != ret_ok)
		xerror(1, "symbol insert fail");
}

void symbol_table_remove_symbol(const char *name)
{
	struct symbol *symbol;
	ret_t ret;
	boolean_t ok;

	return_if_fail(name != NULL);
	return_if_fail(global != NULL);

	ret = hash_table_lookup(top->scope, (void *)name, (void **)&symbol);
	if (ret != ret_ok)
		xerror(1, "lookup fail");

	switch (symbol->type) {
	case SYMBOL_TYPE_ARGUMENT:
		top->nargs--;
		break;
	case SYMBOL_TYPE_VARIABLE:
	case SYMBOL_TYPE_ARRAY:
		top->nvars--;
		break;
	default:
		break;
	}

	ok = hash_table_remove(top->scope, (void *)name);
	if (!ok)
		xerror(1, "symbol remove fail");
}

static struct symbol_table *create_table()
{
	struct symbol_table *table;

	table = xmalloc0(sizeof(*table));
	table->scope = hash_table_new(0, (hash_callback_t)symbol_strhash, (hash_compare_t)symbol_strcmp);
	if (table->scope == NULL)
		xerror(1, "can't create a scope");
	table->ref_count = 1;

	return table;
}

static void table_destroy(struct symbol_table **table)
{
	struct symbol *symbol;
	struct hash_table_iter *iter;
	char *name;

	return_if_fail(table != NULL);
	return_if_fail(*table != NULL);

	iter = hash_table_iterate_init((*table)->scope);
	if (iter == NULL)
		xerror(1, "hash iterator");

	while (hash_table_iterate(iter, (void **)&name, (void **)&symbol))
		symbol_unref(symbol);

	hash_table_iterate_deinit(&iter);
	hash_table_destroy(&(*table)->scope);

	xfree(*table);
	(*table) = NULL;
}

void symbol_table_ref(struct symbol_table *table)
{
	return_if_fail(table != NULL);

	++table->ref_count;
}

void symbol_table_unref(struct symbol_table *table)
{
	return_if_fail(table != NULL);
	return_if_fail(table->ref_count != 0);

	--table->ref_count;

	if (table->ref_count == 0)
		table_destroy(&table);
}

void symbol_ref(struct symbol *symbol)
{
	return_if_fail(symbol != NULL);
	++symbol->ref_count;
}

void symbol_unref(struct symbol *symbol)
{
	return_if_fail(symbol != NULL);
	return_if_fail(symbol->ref_count != 0);

	--symbol->ref_count;
	
	if (symbol->ref_count == 0 && symbol->release != NULL)
		symbol->release(symbol);
}

void symbol_table_create_global_table()
{
	struct symbol_table *table;

	table = create_table();
	global = top = table;
}

struct symbol_table *symbol_table_get_global_table()
{
	return_val_if_fail(global != NULL, NULL);
	return global;
}

/* Push a new namespace scope */
void symbol_table_push()
{
	struct symbol_table *table;

	return_if_fail(global != NULL);

	table = create_table();
	table->prev = top;
	top = table;
}

/* Pop a namespace scope */
void symbol_table_pop()
{
	return_if_fail(global != NULL);
	if (top->prev == NULL)
		xerror(1, "pop global table");

	top = top->prev;
}

void symbol_table_put(struct symbol_table *table)
{
	top = table;
}

struct symbol_table *symbol_table_get_current_table()
{
	return_val_if_fail(top != NULL, NULL);
	return_val_if_fail(global != NULL, NULL);
	return top;
}

struct symbol *symbol_table_lookup_top(const char *name)
{
	struct symbol *symbol;
	ret_t ret;

	return_val_if_fail(name != NULL, NULL);

	if (global == NULL)
		return NULL;

	ret = hash_table_lookup(top->scope, (void *)name, (void **)&symbol);
	if (ret != ret_ok)
		return NULL;

	return symbol;
}

struct symbol *symbol_table_lookup_all(const char *name)
{
	struct symbol *symbol;
	struct symbol_table *table;
	ret_t ret;

	return_val_if_fail(name != NULL, NULL);

	if (global == NULL)
		return NULL;

	symbol = NULL;
	for (table = top; table != NULL; table = table->prev) {
		ret = hash_table_lookup(table->scope, (void *)name, (void**)&symbol);
		if (ret == ret_ok)
			return symbol;
			
	}

	return symbol;
}

